#!/ops/tools/bin/perl
#
##  Copyright (c) 1995-2014 University Corporation for Atmospheric Research
## All rights reserved
#
my $pkgdoc = <<'EOD';
#/**----------------------------------------------------------------------
# @file       raw_C1_vc_splitter.pl
# Given as input one raw downlink file for COSMIC-1 (a large file with all
# VC's combined), split this into VC0,1,2 files.  Ignore VC7 (fill frames)
#
# @author     Doug Hunt
# @since      9/17/2018
# @version    $URL$ $Id$
# @cdaacTask  no
# @usage      raw_C1_vc_splitter.pl trgLv0_2018.249.16.13.006.CUIABA_tlm.gz
# -----------------------------------------------------------------------*/
EOD

$|++;

use lib qw (. /ops/tools/lib);
use strict;
use warnings;
use Getopt::Long;
use RSTools;

if (@ARGV == 0) {
  print $pkgdoc;
  exit -1;
}


# deal with command line options, default is to clean everything.
my $verbose   = 0;
GetOptions (
	    "verbose!"   => \$verbose,
          ) || die "Cannot parse command line options!";

my $inpath = shift;

# Remove spacecraft and ground station headers if requested
# Note that $trig_data is a scalar reference not a scalar
my @opt = (VERBOSE => $verbose);
my $by_vc = split_vc ($inpath, @opt);

foreach my $vc (sort keys %$by_vc) {
  open my $fh, '>', "$inpath.VC$vc" or die "Cannot open output file for VC=$vc";
  print {$fh} $by_vc->{$vc};
  close $fh;
  print "FILE CREATION: $inpath.VC$vc\n";
}

print "Done!\n";

exit;


#/**----------------------------------------------------------------------
# @sub       split_vc
#
# Split one COSMIC-1 'raw' file (telemetry containing multiple Virtual Channels)
# Return the split VC data in a perl hash by VC
#
# @parameter  $infile    Input wrapped multiple-VC file
# @return     $unwrapped Unwrapped instrument data
# @exception  If there is a problem...
# ----------------------------------------------------------------------*/
sub split_vc {

  my $infile = shift;

  my %opt    = (VERBOSE => 0, @_);
  my $verbose = $opt{VERBOSE};

  my $RS_bytes_len      = 160; # length of the Reed-Solomon bytes section
  my $sync              = pack ("N", 0x1ACFFC1D); # CCSDS sync bytes

  my $filetext;
  if ($inpath =~ /\.gz$/) {
    $filetext = `gunzip -c $inpath`;
  } else {
    $filetext = do { local( @ARGV, $/ ) = $inpath; <> } ; # slurp!
  }

  my %appnames = (0x300 => 'GOXA',
                  0x301 => 'GOXB',
                  0x302 => 'TIP',
                  0x303 => 'ATT');

  my %scid = (0x3F1 => 'FM1', 0x3F2 => 'FM2', 0x3F3 => 'FM3',
              0x3F4 => 'FM4', 0x3F5 => 'FM5', 0x3F6 => 'FM6', 0x1ac => 'UNK');

  my %by_vc;     # split data to pass back

  # Match frames one at a time
  my $computed_rs = "\x00" x $RS_bytes_len; # place to put computed RS
  
 FRAME:
  while (1) {
      my $found = ($filetext =~ m{
                                 ($sync)   # sync word
                                 (.{6})    # transfer frame primary header
                                 (.{6})    # packet primary header
                                 (.{7})    # VC0 or VC1 time stamp
                                 (.{1096}) # rest of frame
                                 (.{160})  # Reed Solomon code
                               }msgx);
      last FRAME if (!$found);

      my ($XH_scid, $XH_vcid, $XH_vcfc, $XH_stat) = map { oct "0b$_" } unpack ("x2 a10 a3 x9 a8 a16", unpack ("B*", $2));
      my ($PH_vertype, $PH_apid, $PH_ssc)         = map { oct "0b$_" } unpack ("a4 x a11 x2 a14",     unpack ("B*", $3));

      my $frame  = $2.$3.$4.$5.$6;
      my $rs     = $6;

      RSTools::encode_interleaved($frame, $computed_rs);
      my $rs_err = ($rs ne $computed_rs) ? 1 : 0;
      if ($rs_err) {
	  print "  Reed Solomon error found for VC$XH_vcid\n" if ($verbose);
      }
      printf "%s VC%d %s\n", 
      $scid{$XH_scid}//'UNK', $XH_vcid, $appnames{$PH_apid}//'UNK' if ($verbose);

      $by_vc{$XH_vcid} .= ($sync.$frame.$rs); # data to pass back

  }

  return \%by_vc;

}
